"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = require("assert");
const messageTypes_js_1 = require("../../messageTypes.js");
const index_js_1 = require("../../opLifecycle/index.js");
describe("BatchManager", () => {
    const hardLimit = 950 * 1024;
    const smallMessageSize = 10;
    const generateStringOfSize = (sizeInBytes) => new Array(sizeInBytes + 1).join("0");
    const smallMessage = () => ({
        contents: generateStringOfSize(smallMessageSize),
        type: messageTypes_js_1.ContainerMessageType.FluidDataStoreOp,
    });
    it("BatchManager: 'infinity' hard limit allows everything", () => {
        const message = { contents: generateStringOfSize(1024) };
        const batchManager = new index_js_1.BatchManager({ hardLimit: Infinity, canRebase: true });
        for (let i = 1; i <= 10; i++) {
            assert_1.strict.equal(batchManager.push(message, /* reentrant */ false), true);
            assert_1.strict.equal(batchManager.length, i);
        }
    });
    it("Batch metadata is set correctly", () => {
        const batchManager = new index_js_1.BatchManager({ hardLimit, canRebase: true });
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 0 }, 
        /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 1 }, 
        /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 2 }, 
        /* reentrant */ false), true);
        const batch = batchManager.popBatch();
        assert_1.strict.equal(batch.content[0].metadata?.batch, true);
        assert_1.strict.equal(batch.content[1].metadata?.batch, undefined);
        assert_1.strict.equal(batch.content[2].metadata?.batch, false);
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 0 }, 
        /* reentrant */ false), true);
        const singleOpBatch = batchManager.popBatch();
        assert_1.strict.equal(singleOpBatch.content[0].metadata?.batch, undefined);
    });
    it("Batch content size is tracked correctly", () => {
        const batchManager = new index_js_1.BatchManager({ hardLimit, canRebase: true });
        assert_1.strict.equal(batchManager.push(smallMessage(), /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.contentSizeInBytes, smallMessageSize * batchManager.length);
        assert_1.strict.equal(batchManager.push(smallMessage(), /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.contentSizeInBytes, smallMessageSize * batchManager.length);
        assert_1.strict.equal(batchManager.push(smallMessage(), /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.contentSizeInBytes, smallMessageSize * batchManager.length);
    });
    it("Batch reference sequence number maps to the last message", () => {
        const batchManager = new index_js_1.BatchManager({ hardLimit, canRebase: true });
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 0 }, 
        /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 1 }, 
        /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 2 }, 
        /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.sequenceNumbers.referenceSequenceNumber, 2);
    });
    it("Batch size estimates", () => {
        const batchManager = new index_js_1.BatchManager({ hardLimit, canRebase: true });
        batchManager.push(smallMessage(), /* reentrant */ false);
        // 10 bytes of content + 200 bytes overhead
        assert_1.strict.equal((0, index_js_1.estimateSocketSize)(batchManager.popBatch()), 210);
        for (let i = 0; i < 10; i++) {
            batchManager.push(smallMessage(), /* reentrant */ false);
        }
        // (10 bytes of content + 200 bytes overhead) x 10
        assert_1.strict.equal((0, index_js_1.estimateSocketSize)(batchManager.popBatch()), 2100);
        batchManager.push(smallMessage(), /* reentrant */ false);
        for (let i = 0; i < 9; i++) {
            batchManager.push({
                contents: undefined,
                type: messageTypes_js_1.ContainerMessageType.FluidDataStoreOp,
            }, 
            /* reentrant */ false); // empty op
        }
        // 10 bytes of content + 200 bytes overhead x 10
        assert_1.strict.equal((0, index_js_1.estimateSocketSize)(batchManager.popBatch()), 2010);
    });
    it("Batch op reentry state preserved during its lifetime", () => {
        const batchManager = new index_js_1.BatchManager({ hardLimit, canRebase: true });
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 0 }, 
        /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 1 }, 
        /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 2 }, 
        /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.popBatch().hasReentrantOps, false);
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 0 }, 
        /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 1 }, 
        /* reentrant */ true, 
        /* currentClientSequenceNumber */ undefined), true);
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 2 }, 
        /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.popBatch().hasReentrantOps, true);
        assert_1.strict.equal(batchManager.push({ ...smallMessage(), referenceSequenceNumber: 0 }, 
        /* reentrant */ false), true);
        assert_1.strict.equal(batchManager.popBatch().hasReentrantOps, false);
    });
});
//# sourceMappingURL=batchManager.spec.js.map